//*****************************************************************************

//
// Matlab C routine file:  chain
//
// Stores a silhouette as a chained sequence of turns.
//
// Written 6/02 by N. Howe.
// Updated 8/02 by N. Howe.
//

//*****************************************************************************


#include "mex.h"

#define NORTH 0
#define SOUTH 1
#define EAST 2
#define WEST 3

//*****************************************************************************


//
//  Adds bits to the buffer.
//

__inline void emit_bits(int bits, int nbit, unsigned char **bptr, int *bbit) {
  int val = (**bptr)+(bits<<(*bbit));

  (*bbit) += nbit;
  if (*bbit < 8) {
    (**bptr) = val;
  } else {
    (**bptr) = val%256;
    //mexPrintf("Byte:  %d.\n",(**bptr));
    (*bptr)++;
    (**bptr) = val/256;
    (*bbit) -= 8;
  }
  //mexPrintf("Emitted %d bits (value %d) ending at byte %d, position %d.\n",nbit,bits,*bptr,*bbit);
}
// end of emit_bits()

//*****************************************************************************


//
//  Gateway driver to call the calculation from Matlab.
//

// This is the Matlab entry point
void 
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  int i, j, i0, j0;
  int last;
  bool solid;
  int nrow, ncol, jnrow, jncol, bbit, start, nbyte;
  int *jarr;
  unsigned char *buf, *bptr, *res;
  double *arr, *p1, *p2, *p3, *p4;
	
  // check for proper number of arguments
  if (nrhs != 1) {
    mexErrMsgTxt("Exactly one argument required.");
  }
  if (nlhs > 2) {
    mexErrMsgTxt("Too many output arguments.");
  }

  // check format of arguments
  if (!mxIsDouble(prhs[0])||mxIsComplex(prhs[0])) {
    mexErrMsgTxt("Inputs must be type double.");
  }
  nrow = mxGetM(prhs[0]);
  ncol = mxGetN(prhs[0]);
  arr = mxGetPr(prhs[0]);
  jnrow = nrow+1;
  jncol = ncol+1;
  //mexPrintf("Dimensions are %d by %d.\n",nrow,ncol);
  //mxAssert(((nrow <= 256)&&(ncol <= 256)),"Dimensions must be less than 256 by 256.");

  // allocate space
  jarr = (int*)mxMalloc(jnrow*jncol*sizeof(int));
  buf = (unsigned char*)mxMalloc(jnrow*jncol*sizeof(unsigned char)/2+8);
  //if (nlhs > 1) {
  //  plhs[1] = mxCreateDoubleMatrix(1, 16, mxREAL);
  //  jcnt = mxGetPr(plhs[1]);
  //  for (i = 0; i < 16; i++) {
  //    jcnt[i] = 0;
  //  }
  //}

  // compute junctions
  //mexPrintf("Joint array transpose:\n");
  //mexPrintf("Size of joint array:  %dx%d.\n",jnrow,jncol);
  p1 = arr;
  p2 = arr+1;
  p3 = arr+nrow;
  p4 = arr+nrow+1;
  // center section:
  for (j = 1; j < jncol-1; j++) {
    for (i = 1; i < jnrow-1; i++) {
      jarr[j*jnrow+i] = ((*p1!=0) != (*p2!=0))+2*((*p1!=0) != (*p3!=0))
        +4*((*p3!=0) != (*p4!=0))+8*((*p2!=0) != (*p4!=0));
      //mexPrintf("%d ",jarr[j*jnrow+i]);
      //mexPrintf("Joint array at (%d,%d):  %d.\n",i,j,jarr[j*jnrow+i]);
      p1++;
      p2++;
      p3++;
      p4++;
    }
    //mexPrintf("\n");
    p1++;	// skip last row
    p2++;
    p3++;
    p4++;
  }
  // edges:
  p2 = arr;
  p4 = arr+nrow;
  p1 = arr+nrow-1;
  p3 = arr+2*nrow-1;
  for (j = 1; j < jncol-1; j++) {
    jarr[j*jnrow] = (0 != (*p2!=0))+4*(0 != (*p4!=0))+8*((*p2!=0) != (*p4!=0));
    jarr[j*jnrow+jnrow-1] = ((*p1!=0) != 0)+2*((*p1!=0) != (*p3!=0))
      +4*((*p3!=0) != 0);
    p1 += nrow;
    p2 += nrow;
    p3 += nrow;
    p4 += nrow;
  }
  p3 = arr;
  p4 = arr+1;
  p1 = arr+(ncol-1)*nrow;
  p2 = arr+(ncol-1)*nrow+1;
  for (i = 1; i < jnrow-1; i++) {
    jarr[i] = 2*(0 != (*p3!=0))+4*((*p3!=0) != (*p4!=0))+8*(0 != (*p4!=0));
    jarr[(jncol-1)*jnrow+i] = ((*p1!=0) != (*p2!=0))+2*((*p1!=0) != 0)
      +8*((*p2!=0) != 0);
    p1++;
    p2++;
    p3++;
    p4++;
  }
  jarr[0] = ((arr[0]) ? 12:0);
  jarr[jnrow-1] = ((arr[nrow-1]) ? 6:0);
  jarr[jnrow*(jncol-1)] = ((arr[(ncol-1)*nrow]) ? 9:0);
  jarr[jnrow*(jncol-1)+jnrow-1] = ((arr[(ncol-1)*nrow+nrow-1]) ? 3:0);
  //mexPrintf("Junctions.\n");

  // compute bits
  for (start = 0; (jarr[start] == 0)&&(start < jnrow*jncol); start++);
  if (start >= jnrow*jncol) {
    // nothing to encode; emit empty result
    mexWarnMsgTxt("Encoding empty silhouette.");
    i = i0 = 0;
    j = j0 = 0;
    bbit = 0;
    bptr = buf;
  } else {
    //mexPrintf("Start at %d.\n",start);
    bbit = 0;
    bptr = buf;
    (*bptr) = 0;

    i = i0 = start%jnrow;
    j = j0 = start/jnrow;
    emit_bits(EAST,2,&bptr,&bbit);
    j++;
    last = EAST;
  }
  while ((i != i0)||(j != j0)) {
    // solid refers to SE pixel of junction
    if ((i >= nrow)||(j >= ncol)) {
      solid = 0;
    } else {
      solid = (arr[i+j*nrow] != 0);
    }

    //mexPrintf("Loop: (%d,%d): %d (solid? %d).\n",i,j,jarr[j*jnrow+i],solid);
    mxAssert(((i >= 0)&&(j >= 0)&&(i < jnrow)&&(j < jncol)),"Out of bounds.");
    switch (jarr[j*jnrow+i]) {
    case 3:
      if (solid) {
        // E turning N
        i--;
        last = NORTH;
      } else {
        // S turning W
        j--;
        last = WEST;
      }
      break;
    case 5:
      if (solid) {
        // E continuing E
        j++;
      } else {
        // W continuing W
        j--;
      }
      break;
    case 6:
      if (solid) {
        // S turning E
        j++;
        last = EAST;
      } else {
        // W turning N
        i--;
        last = NORTH;
      }
      break;
    case 9:
      if (solid) {
        // N turning W
        j--;
        last = WEST;
      } else {
        // E turning S
        i++;
        last = SOUTH;
      }
      break;
    case 10:
      if (solid) {
        // N continuing N
        i--;
      } else {
        // S continuing S
        i++;
      }
      break;
    case 12:
      if (solid) {
        // N turning E
        j++;
        last = EAST;
      } else {
        // W turning S
        i++;
        last = SOUTH;
      }
      break;
    case 15:
      if (solid) {
        if (last == SOUTH) {
          // S turning E
          j++;
          last = EAST;
        } else {
          // N turning W
          j--;
          last = WEST;
        }
      } else {
        if (last == EAST) {
          // E turning N
          i--;
          last = NORTH;
        } else {
          // W turning S
          i++;
          last = SOUTH;
        }
      }
      break;
    }
    //    }
    emit_bits(last,2,&bptr,&bbit);
  }
  //mexPrintf("Done:  (%d,%d): %d (solid? %d).\n",i,j,jarr[j*jnrow+i],solid);

  // pad out last bits if necessary, for cmpChains
  if (bbit == 4) {
    emit_bits(EAST,2,&bptr,&bbit);
    emit_bits(WEST,2,&bptr,&bbit);
  }

  // prepare output
  nbyte = (bptr-buf+(bbit+7)/8);
  //mexPrintf("Result is %d bytes (%d full + %d bits).",nbyte,bptr-buf,bbit);
  plhs[0] = mxCreateNumericArray(1, &nbyte, mxUINT8_CLASS, mxREAL);
  res = (unsigned char*)mxGetPr(plhs[0]);
  for (i = 0; i < nbyte; i++) {
    res[i] = buf[i];
  }
  if (nlhs > 1) {
    // output junctions
    plhs[1] = mxCreateDoubleMatrix(jnrow,jncol, mxREAL);
    arr = mxGetPr(plhs[1]);
    for (j = 0; j <jncol; j++) {
      for (i = 0; i < jnrow; i++) {
        arr[j*jnrow+i] = jarr[j*jnrow+i];
      }
    }
  }

  // free stuff
  mxFree(jarr);
  mxFree(buf);
}

//*****************************************************************************
