//***************************************************************************/
//
// Matlab C routine file:  chamferDistance
//
// Program to quickly compare chained silhouettes with the edge distance
// transform of another.  Results are sum of dtrans distances.
//
// Written 10/03 by N. Howe.  Updated 7/04.
//
// Some notes:
// 1.  Computes half chamfer distance => use on outlines of silhouettes.
// 2.  The chained shape is assumed to be centered within the distance
//     transform matrix provided.
// 3.  Results inaccurate if chained silhouette extends beyond edge of
//     distance transform matrix.
//
//***************************************************************************/

#include <math.h>
#include "mex.h"
#include "chaincodes.h"

//***************************************************************************/

#define HALFSQRT 0.7071067811865476

//***************************************************************************/

#define ABS(a) (((a) < 0) ? -(a) : (a))
// macro for computing absolute value

#define ROUND(a) ((int)((a)+0.5))
// macro for rounding

#define BOUND(v, l, h) (((v) > (l)) ? (((v) < (h)) ? (v) : (h)) : (l))
// macro for bounding a number by two others

#define errCheck(a,b) if (!(a)) mexErrMsgTxt((b));
// macro for argument checking

//*****************************************************************************
//
//  Returns the dimensions of a chained representation.
//

void chainDim(const mxArray *ch, int &nrow, int &ncol, int &roff, int &coff) {
  int r, c, b;
  int nbyte, rmin = 0, rmax = 0, cmin = 0, cmax = 0;
  unsigned char *buf;

  r = c = 0;
  nbyte = mxGetNumberOfElements(ch);
  buf = (unsigned char *)mxGetPr(ch);

  for (b = 0; b < nbyte; b++) {
    r += -dy[buf[b]];
    c += dx[buf[b]];
    if (r < rmin) {
      rmin = r;
    }
    if (r > rmax) {
      rmax = r;
    }
    if (c < cmin) {
      cmin = c;
    }
    if (c > cmax) {
      cmax = c;
    }
  }
  nrow = rmax-rmin;
  ncol = cmax-cmin;
  roff = -rmin;
  coff = -cmin;
} // end of chainDim()

//*****************************************************************************
//
//  Computes the sum chamfer distance on the chain.
//

double chainChamfer(const mxArray *ch, const mxArray *dt, double scale) {
  int r, c, b, rs, cs;
  int nbyte, m1, n1, m2, n2;
  double total, roff, coff;
  unsigned char *buf;
  double *dp;
  
  nbyte = mxGetNumberOfElements(ch);
  buf = (unsigned char *)mxGetPr(ch);
  chainDim(ch,m1,n1,r,c);
  m2 = mxGetM(dt);
  n2 = mxGetN(dt);
  dp = mxGetPr(dt);
  roff = (m2-m1*scale)/2;
  coff = (n2-n1*scale)/2;

  if ((m1*scale>m2)|(n1*scale>n2)) {
    mexWarnMsgTxt(
      "Distance transform matrix is too small:  result will be inaccurate.");
  }

  b = 0;
  rs = (int)BOUND(r*scale+roff,0,m2-1);
  cs = (int)BOUND(c*scale+coff,0,n2-1);
  mxAssert(rs+m2*cs < mxGetNumberOfElements(prhs[1]),
           "Computed index out of bounds.");
  total = dp[rs+m2*cs];
  for (b = 0; b < nbyte; b++) {
    r += -dy[buf[b]];
    c += dx[buf[b]];
    rs = (int)BOUND(r*scale+roff,0,m2-1);
    cs = (int)BOUND(c*scale+coff,0,n2-1);
    mxAssert(rs+m2*cs < mxGetNumberOfElements(prhs[1]),
             "Computed index out of bounds.");
    total += dp[rs+m2*cs];
  }

  return total;
}  // end of chainChamfer()

//*****************************************************************************
//
//  mexFunction() is a gateway driver to call the code from Matlab.
//  This is the Matlab entry point.
//
//***************************************************************************/

void 
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  int i, nch, nsc;
  double cscale, defscale = 1, *scale, *arr;
  const mxArray *cell;
  
  // check for proper number of arguments
  errCheck(nrhs >= 2,"Must provide chained silhouette and dtrans.");
  errCheck(nrhs <= 3,"Too many input arguments.");
  errCheck(nlhs <= 1,"Too many output arguments.");
  
  // check inputs
  if (mxIsCell(prhs[0])) {
    nch = mxGetNumberOfElements(prhs[0]);
  } else {
    nch = 1;
  }
  errCheck(mxIsDouble(prhs[1]),"Second input must be double matrix.");
  if (nrhs > 2) {
    errCheck(mxIsDouble(prhs[2]),"Third input must be double.");
    nsc = mxGetNumberOfElements(prhs[2]);
    errCheck((nsc == nch)||(nsc == 1),
             "Third input must be scalar or same size as first.");
    scale = mxGetPr(prhs[2]);
  } else {
    nsc = 1;
    scale = &defscale;
  }

  // prepare output
  plhs[0] = mxCreateDoubleMatrix(1,nch,mxREAL);
  arr = mxGetPr(plhs[0]);

  // compute
  for (i = 0; i < nch; i++) {
    if (mxIsCell(prhs[0])) {
      cell = mxGetCell(prhs[0],i);
    } else {
      cell = prhs[0];
    }
    if (nsc > 1) {
      cscale = scale[i];
    } else {
      cscale = *scale;
    }

    errCheck(mxIsUint8(cell),
             "First input must be uint8 vector or cell array of same.");
    arr[i] = chainChamfer(cell,prhs[1],cscale);
  }
}
// end of main()

//*****************************************************************************
