//***************************************************************************/
//
// Matlab C routine file:  chamferView
//
// Returns visual representation of chamferDistance computation.
//
// Written 10/03 by N. Howe.  Updated 7/04.
//
//***************************************************************************/

#include <math.h>
#include "mex.h"
#include "chaincodes.h"

//***************************************************************************/

#define HALFSQRT 0.7071067811865476

//***************************************************************************/

#define ABS(a) (((a) < 0) ? -(a) : (a))
// macro for computing absolute value

#define ROUND(a) ((int)((a)+0.5))
// macro for rounding

#define BOUND(v, l, h) (((v) > (l)) ? (((v) < (h)) ? (v) : (h)) : (l))
// macro for bounding a number by two others

#define errCheck(a,b) if (!(a)) mexErrMsgTxt((b));
// macro for argument checking

//*****************************************************************************
//
//  NaN() returns IEEE double representation of NaN.
//

inline double NaN() {
  //unsigned long nana[2];
  //nan[0]=0xffffffff;
  //nan[1]=0x7fffffff;
  static unsigned long nana[2]={0xffffffff, 0x7fffffff};
  return (*(double*)nana);
} // end of NaN()


//*****************************************************************************
//
//  Returns the dimensions of a chained representation.
//

void chainDim(const mxArray *ch, int &nrow, int &ncol, int &roff, int &coff) {
  int r, c, b;
  int nbyte, rmin = 0, rmax = 0, cmin = 0, cmax = 0;
  unsigned char *buf;

  r = c = 0;
  nbyte = mxGetNumberOfElements(ch);
  buf = (unsigned char *)mxGetPr(ch);

  for (b = 0; b < nbyte; b++) {
    r += -dy[buf[b]];
    c += dx[buf[b]];
    if (r < rmin) {
      rmin = r;
    }
    if (r > rmax) {
      rmax = r;
    }
    if (c < cmin) {
      cmin = c;
    }
    if (c > cmax) {
      cmax = c;
    }
  }
  nrow = rmax-rmin;
  ncol = cmax-cmin;
  roff = -rmin;
  coff = -cmin;
} // end of chainDim()

//*****************************************************************************
//
//  Computes the sum chamfer distance on the chain.
//

void chainChamfer(const mxArray *ch, const mxArray *dt, double scale, double *arr) {
  int r, c, b, rs, cs;
  int nbyte, m1, n1, m2, n2;
  double total, roff, coff;
  unsigned char *buf;
  double *dp;
  
  nbyte = mxGetNumberOfElements(ch);
  buf = (unsigned char *)mxGetPr(ch);
  chainDim(ch,m1,n1,r,c);
  m2 = mxGetM(dt);
  n2 = mxGetN(dt);
  dp = mxGetPr(dt);
  roff = (m2-m1*scale)/2;
  coff = (n2-n1*scale)/2;
  //mexPrintf("Dimensions:  %gx%g vs. %dx%d.  Start:  (%d,%d).  Off:  (%d,%d)\n",
  //          m1*scale,n1*scale,m2,n2,r,c,roff,coff);

  if ((m1*scale>m2)|(n1*scale>n2)) {
    mexWarnMsgTxt(
      "Distance transform matrix is too small:  result will be inaccurate.");
  }

  //scale = sqrt((m2*n2)/(m1*n1));
  b = 0;
  rs = (int)BOUND(r*scale+roff,0,m2-1);
  cs = (int)BOUND(c*scale+coff,0,n2-1);
  //mexPrintf("Start:  (%d,%d) -> (%d,%d).\n",b,r,c,rs,cs);
  mxAssert(rs+m2*cs < mxGetNumberOfElements(dt),
           "Computed index out of bounds.");
  total = dp[rs+m2*cs];
  for (b = 0; b < nbyte; b++) {
    //mexPrintf("Position:  (%d,%d).  Motion:  (%d,%d)\n",r,c
    //          ,-dy[buf[b]],dx[buf[b]]);
    r += -dy[buf[b]];
    c += dx[buf[b]];
    rs = (int)BOUND(r*scale+roff,0,m2-1);
    cs = (int)BOUND(c*scale+coff,0,n2-1);
    //mexPrintf("Byte %d:  (%d,%d) -> (%d,%d).\n",b,r,c,rs,cs);
    mxAssert(rs+m2*cs < mxGetNumberOfElements(dt),
             "Computed index out of bounds.");
    arr[rs+m2*cs] = dp[rs+m2*cs];
    }
}  // end of chainChamfer()

//*****************************************************************************
//
//  mexFunction() is a gateway driver to call the code from Matlab.
//  This is the Matlab entry point.
//
//***************************************************************************/

void 
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  int i, nch;
  double scale, *arr;
  const mxArray *cell;
  
  // check for proper number of arguments
  errCheck(nrhs >= 2,"Must provide chained silhouette and dtrans.");
  errCheck(nrhs <= 3,"Too many input arguments.");
  errCheck(nlhs <= 1,"Too many output arguments.");
  
  // check input
  if (mxIsCell(prhs[0])) {
    nch = mxGetNumberOfElements(prhs[0]);
    mexErrMsgTxt("Only implemented for single images.");
  } else {
    nch = 1;
  }
  if (!mxIsDouble(prhs[1])||mxIsComplex(prhs[1])) {
    mexErrMsgTxt("Second input must be type double real.");
  }
  errCheck(mxIsDouble(prhs[1]),"Second input must be double matrix.");
  if (nrhs > 2) {
    errCheck(mxIsDouble(prhs[2])|(mxGetNumberOfElements(prhs[2]) != 1),
             "Third input must be double scalar.");
    scale = *(mxGetPr(prhs[2]));
  } else {
    scale = 1;
  }

  // prepare output
  plhs[0] = mxCreateDoubleMatrix(mxGetM(prhs[1]),mxGetN(prhs[1]),mxREAL);
  arr = mxGetPr(plhs[0]);
  for (i = 0; i < mxGetNumberOfElements(prhs[1]); i++) {
    arr[i] = NaN();
  }

  // compute
  for (i = 0; i < nch; i++) {
    chainChamfer(prhs[0],prhs[1],scale,arr);
  }
}
// end of main()

//*****************************************************************************
