//*****************************************************************************

//
// Matlab C routine file:  taCompare
//
// Written 6/02 by N. Howe.
// Modified 8/02 by N. Howe.
//
// Compares chains using the turning angle.
//

//*****************************************************************************

#include <math.h>
#include "mex.h"
#include "chaincodes.h"

//*****************************************************************************

#define NORTH 0
#define SOUTH 1
#define EAST 2
#define WEST 3

//*****************************************************************************

#define MIN(a, b) (((a) < (b)) ? (a) : (b))
// macro for computing minimum

#define ABS(x) (((x) < 0) ? (-(x)):(x))
// macro for taking absolute value

#define ADDMOD(x,y,m) (((x)+(y) < (m)) ? (((x)+(y) >= 0) ? (x)+(y):(x)+(y)+(m)) : (x)+(y)-(m))
// macro for computing a modular sum

//*****************************************************************************

const double PI = 3.1415926535897931;
const double TWOPI = 6.2831853071795862;

//*****************************************************************************

//
//  Compute the new theta relative to the old.
//  Necessary in case of spirals, etc.
//
//  R/O thold:  last theta mod 2*pi
//  R/O code:   code of current segment
//

__inline double dtheta(double mthold, int code) {
  double thnew = theta[code];
  double dth;

  if (mthold < thnew) {
    if (thnew-mthold < PI) {
      // new
      //  | / old
      //  |/
      dth = thnew-mthold;
    } else if (thnew-mthold > PI) {
      //    / old
      //   /
      //  |
      //  |
      // new
      dth = thnew-(mthold+TWOPI);
    } else if (convex[code]) {
      dth = thnew-mthold;
      //mexPrintf("Convex, lt.");
    } else if (concave[code]) {
      dth = thnew-(mthold+TWOPI);
      //mexPrintf("Concave, lt.");
    } else {
      mexErrMsgTxt("Self-crossing chain?");
    }
  } else {
    if (mthold-thnew < PI) {
      // old
      //  | / new
      //  |/
      dth = thnew-mthold;
    } else if (mthold-thnew > PI) {
      //    / new
      //   /
      //  |
      //  |
      // old
      dth = thnew+TWOPI-mthold;
    } else if (convex[code]) {
      //mexPrintf("Convex, gt.");
      dth = thnew+TWOPI-mthold;
    } else if (concave[code]) {
      dth = thnew-mthold;
      //mexPrintf("Concave, gt.");
    } else {
      mexErrMsgTxt("Self-crossing chain?");
    }
  }
  return dth;
}
// end of dtheta()

//*****************************************************************************

//
//  Gateway driver to call the calculation from Matlab.
//

// This is the Matlab entry point
void 
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  int i, j, c1, c2;
  int n1, n2, ch1n, ch2n;
  unsigned char *ch1data, *ch2data;
  double th1, th2, mth1, mth2, l1, l2, l, d, dl, dl1, dl2, dth;
  double *out, *len1, *len2;
  const mxArray *ch1, *ch2;
	
  // check for proper number of arguments
  if ((nrhs != 2)&&(nrhs != 4)) {
    mexErrMsgTxt("Exactly two or four arguments required.");
  }
  if (nlhs > 2) {
    mexErrMsgTxt("Too many output arguments.");
  }

  // check format of arguments
  if (mxIsCell(prhs[0])) {
    n1 = mxGetNumberOfElements(prhs[0]);
  } else {
    n1 = 1;
  }
  if (mxIsCell(prhs[1])) {
    n2 = mxGetNumberOfElements(prhs[1]);
  } else {
    n2 = 1;
  }

  if (nrhs == 4) {
    // read provided lengths
    if (!mxIsDouble(prhs[2])||mxIsComplex(prhs[2])) {
      mexErrMsgTxt("Length inputs must be type double.");
    }
    if (mxGetNumberOfElements(prhs[2]) != n1) {
      mexErrMsgTxt("Length of third argument must match first.");
    }
    len1 = mxGetPr(prhs[2]);
    if (!mxIsDouble(prhs[3])||mxIsComplex(prhs[3])) {
      mexErrMsgTxt("Length inputs must be type double.");
    }
    if (mxGetNumberOfElements(prhs[3]) != n2) {
      mexErrMsgTxt("Length of fourth argument must match second.");
    }
    len2 = mxGetPr(prhs[3]);
  } else {
    // calculate lengths
    //mexPrintf("Computing lengths.\n");
    len1 = (double*)mxMalloc(n1*sizeof(double));
    len2 = (double*)mxMalloc(n2*sizeof(double));
    for (c1 = 0; c1 < n1; c1++) {
      if (mxIsCell(prhs[0])) {
        ch1 = mxGetCell(prhs[0],c1);
      } else {
        ch1 = prhs[0];
      }
      if (!mxIsUint8(ch1)) {
        mexErrMsgTxt("Inputs must be type uint8 or cell array of uint8.");
      }
      ch1data = (unsigned char *)(mxGetPr(ch1));
      len1[c1] = 0;
      for (i = 0; i < mxGetNumberOfElements(ch1); i++) {
        len1[c1] += len[ch1data[i]];
      }
      //mexPrintf("Length of input1(%d) = %f.\n",c1,len1[c1]);
    }
    for (c2 = 0; c2 < n2; c2++) {
      if (mxIsCell(prhs[1])) {
        ch2 = mxGetCell(prhs[1],c2);
      } else {
        ch2 = prhs[1];
      }
      if (!mxIsUint8(ch2)) {
        mexErrMsgTxt("Inputs must be type uint8 or cell array of uint8.");
      }
      ch2data = (unsigned char *)(mxGetPr(ch2));
      len2[c2] = 0;
      for (j = 0; j < mxGetNumberOfElements(ch2); j++) {
        len2[c2] += len[ch2data[j]];
      }
      //mexPrintf("Length of input2(%d) = %f.\n",c2,len2[c2]);
    }
  }
  //mexPrintf("Arguments checked.\n");
  
  // create output
  plhs[0] = mxCreateDoubleMatrix(n1, n2, mxREAL);
  out = mxGetPr(plhs[0]);

  // compare bits
  for (c1 = 0; c1 < n1; c1++) {
    if (mxIsCell(prhs[0])) {
      ch1 = mxGetCell(prhs[0],c1);
    } else {
      ch1 = prhs[0];
    }
    for (c2 = 0; c2 < n2; c2++) {
      if (mxIsCell(prhs[1])) {
        ch2 = mxGetCell(prhs[1],c2);
      } else {
        ch2 = prhs[1];
      }
      ch1data = (unsigned char *)(mxGetPr(ch1));
      ch2data = (unsigned char *)(mxGetPr(ch2));
      i = j = 0;
      th1 = 0;
      th2 = 0;
      mth1 = 0;
      mth2 = 0;
      l1 = 0;  // progress around perimeter of chain 1
      l2 = 0;  // progress around perimeter of chain 2
      l = 0;   // overall progress
      d = 0;
      ch1n = mxGetNumberOfElements(ch1);
      ch2n = mxGetNumberOfElements(ch2);
      if ((ch1n == 0)||(ch2n == 0)) {
        out[c1+n1*c2] = mxGetNaN();
      } else {
        while ((i < ch1n)||(j < ch2n)) {
          //mexPrintf("Comp: ch1 @ %d, pct %f; ch2 @ %d, pct %f; all %f.\n"
          //          ,i,l1,j,l2,l);
          //mexPrintf("Choice:  l1 = %f, l2 = %f, i = %d, j = %d.\n",l1,l2,i,j);
          if ((l1 < l2)||(j == ch2n)) {
            // taking a step on ch1
            mxAssert(i < ch1n,"Exceeded length of first chain.");
            //mexPrintf("Step on 1 (#%d).  Code = %d; len = %f; theta = %f.\n"
            //          ,i,ch1data[i],len[ch1data[i]],theta[ch1data[i]]);
            dl1 = len[ch1data[i]]/len1[c1];
            dl = MIN(dl1,l2-l1);
            dth = dtheta(mth1,ch1data[i]);
            th1 += dth;
            //mexPrintf("%f; convex-%d concave-%d\n"
            //          ,th1,convex[ch1data[i]],concave[ch1data[i]]);
            mth1 = theta[ch1data[i]];
            l1 += dl1;
            l += dl;
            //mexPrintf("%f + %f*%f = %f.\n"
            //          ,d,dl,ABS(th1-th2),dl*ABS(th1-th2));
            d += dl*ABS(th1-th2);
            //mexPrintf("Altered theta by %f to get %f (%f modular).\n"
            //          ,dth,th1,theta[ch1data[i]]);
            //mexPrintf("Stepped %f; angular distance %f.  (%f vs. %f)\n"
            //          ,dl,ABS(th1-th2),th1,th2);
            i++;
          } else {
            // taking a step on ch2
            //mexPrintf("Step on 2 (#%d).  Code = %d; len = %f; theta = %f.\n"
            //          ,j,ch2data[j],len[ch2data[j]],theta[ch2data[j]]);
            dl2 = len[ch2data[j]]/len2[c2];
            dl = MIN(dl2,l1-l2);
            dth = dtheta(mth2,ch2data[j]);
            th2 += dth;
            mth2 = theta[ch2data[j]];
            //mth2 = ADDMOD(mth2,dth,TWOPI);
            l2 += dl2;
            l += dl;
            //mexPrintf("%f + %f*%f = %f.\n"
            //          ,d,dl,ABS(th1-th2),dl*ABS(th1-th2));
            d += dl*ABS(th1-th2);
            //mexPrintf("Altered theta by %f to get %f (%f modular).\n"
            //          ,dth,th2,theta[ch2data[j]]);
            //mexPrintf("Stepped %f; angular distance %f.  (%f vs. %f)\n"
            //          ,dl,ABS(th1-th2),th1,th2);
            j++;
          }
        }
        //mexPrintf("Comp: ch1 @ %d, pct %f; ch2 @ %d, pct %f; all %f.\n"
        //          ,i,l1,j,l2,l);
        out[c1+n1*c2] = d;
      }
    }
  }
}

//*****************************************************************************
