//*****************************************************************************

//
// Matlab C routine file:  unchain
//
// Restores a silhouette from a chained sequence of turns.
//
// Written 6/02 by N. Howe.
// Modified 8/02 by N. Howe.
//

//*****************************************************************************

#include "mex.h"

#define NORTH 0
#define SOUTH 1
#define EAST 2
#define WEST 3

#define MIN(x,y) (((x) <= (y)) ? (x) : (y))

//*****************************************************************************

//
//  Gets bits from the buffer.
//

__inline int get_bits(int nbit, unsigned char **bptr, int *bbit) {
  int res;
  int n,d;
	
  //mexPrintf("Buffer:  [...%d, %d, ...]; shifted:  %d; modulated:  %d.\n",(*bptr)[0],(*bptr)[1],(*bptr)[0]>>(*bbit),((**bptr)>>(*bbit))%(1<<nbit));
  if ((*bbit)+nbit < 8) {
    //res = ((**bptr)>>(*bbit))%(1<<nbit)
    // weird problems happening here.  Using extra variables
    // seems to make it work.
    n = (*bptr)[0]>>(*bbit);
    d = (1<<nbit);
    res = n%d;
    //mexPrintf("%d,%d.\n",res,((**bptr)>>(*bbit))%(1<<nbit));
    //mexPrintf("%d,%d,%d.\n",res,((**bptr)>>(*bbit))%(1<<nbit),res = ((**bptr)>>(*bbit))%(1<<nbit));
    //mexPrintf("  Got %d from %d bits of %d at %d.\n",res,nbit,**bptr,*bbit);
    (*bbit) += nbit;
  } else {
    //res = (((**bptr)+256*((*bptr)[1]))>>(*bbit))%(1<<nbit);
    // weird problems happening here.  Using extra variables
    // seems to make it work.
    n = ((*bptr)[0]+256*((*bptr)[1]))>>(*bbit);
    d = (1<<nbit);
    res = n%d;
    //mexPrintf("Chewed byte %d (value %d).\n",*bptr,**bptr);
    (*bptr)++;
    (*bbit) = ((*bbit)+nbit)%8;
  }
  //mexPrintf("Read value %d from %d bits ending at byte %d, position %d.\n",res,nbit,*bptr,*bbit);
  return res;
}
// end of get_bits()

//*****************************************************************************

//
//  Reads variable-length codes from the buffer.
//  Codes are two bits in length.
//

__inline int get_code(unsigned char **bptr, int *bbit) {
  int res;

  res = get_bits(2,bptr,bbit);
  if (res == 3) {
    res += 4*get_bits(2,bptr,bbit);
  }
  //mexPrintf("Got code %d from %d bits.\n",res,((res > 3) ? 4:2));
  return res;
}
// end of get_code()

//*****************************************************************************

//
//  Gateway driver to call the calculation from Matlab.
//

// This is the Matlab entry point
void 
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  int i, j, dir, last;
  bool inside;
  int nbyte, nrow, ncol, jnrow, jncol, bbit;
  int jrmin, jcmin,jrmax,jcmax;
  int *jarr;
  unsigned char *buf, *bptr;
  double *arr, *jbds;
	
  // check for proper number of arguments
  if (nrhs < 1) {
    mexErrMsgTxt("At least one argument required.");
  }
  if (nrhs > 2) {
    mexErrMsgTxt("At most two arguments allowed.");
  }
  if (nlhs > 2) {
    mexErrMsgTxt("Too many output arguments.");
  }

  // check format of arguments
  if (!mxIsNumeric(prhs[0])||mxIsDouble(prhs[0])) {
    mexErrMsgTxt("Input must be type uint8.");
  }
  nbyte = mxGetNumberOfElements(prhs[0]);
  buf = (unsigned char *)mxGetPr(prhs[0]);
  //mexPrintf("Unchaining %d bytes.\n",nbyte);
  //for (i = 0; i < nbyte; i++) {
  //  mexPrintf("%d ",buf[i]);
  //}
  //mexPrintf("\n");

  if (nrhs == 2) {
    if (!mxIsDouble(prhs[1])||mxIsComplex(prhs[1])) {
      mexErrMsgTxt("Second input must be type double.");
    }
    if (mxGetNumberOfElements(prhs[1]) != 4) {
      mexErrMsgTxt("Second input must consist of [rmin, cmin, rmax, cmax].");
    }
    jbds = mxGetPr(prhs[1]);
    jrmin = (int)(jbds[0]);
    jcmin = (int)(jbds[1]);
    jrmax = (int)(jbds[2]);
    jcmax = (int)(jbds[3]);
  } else if (nbyte == 0) {
    jrmin = 0;
    jcmin = 0;
    jrmax = 0;
    jcmax = 0;
  } else {
    // compute silhouette size
    jrmin = jcmin = jrmax = jcmax = 0;
    bptr = buf;
    bbit = 0;
    dir = get_bits(2,&bptr,&bbit);	// always EAST
    //mexPrintf("Initial direction: %d.\n",dir);
    mxAssert((dir == EAST),"Initial direction not East.");
    i = 0;
    j = 1;
    while ((i != 0)||(j != 0)) {
      mxAssert((bptr-buf < nbyte),"Overran buffer.");
      dir = get_bits(2,&bptr,&bbit);
      //mexPrintf("Loc:  (%d,%d); dir = %d.\n",i,j,dir);
      switch (dir) {
      case NORTH:
        i--;
        break;
      case SOUTH:
        i++;
        break;
      case EAST:
        j++;
        break;
      case WEST:
        j--;
        break;
      }
      if (i < jrmin) {
        jrmin = i;
      }
      if (j < jcmin) {
        jcmin = j;
      }
      if (i > jrmax) {
        jrmax = i;
      }
      if (j > jcmax) {
        jcmax = j;
      }
    }
  }

  jnrow = (jrmax-jrmin)+1;
  jncol = (jcmax-jcmin)+1;
  nrow = jnrow+1;
  ncol = jncol+1;
  //mexPrintf("Bounds: rows %d - %d; cols %d - %d.\n",jrmin,jrmax,jcmin,jcmax);

  // allocate space
  jarr = (int*)mxMalloc(jnrow*jncol*sizeof(int));
  for (i = 0; i < jnrow*jncol; i++) {
    jarr[i] = 0;
  }

  // decode bits
  if (nbyte > 0) {
    i = -jrmin;
    j = -jcmin;
    bptr = buf;
    bbit = 0;
    dir = get_bits(2,&bptr,&bbit);  // always EAST
    jarr[j*jnrow+i] += 12;
    last = EAST;
    j++;
    while ((i != -jrmin)||(j != -jcmin)) {
      mxAssert((i >= 0)&&(j >= 0)&&(i < jnrow)&&(j < jncol),"Out of bounds.");
      dir = get_bits(2,&bptr,&bbit);
      switch (dir) {
      case NORTH:
        switch (last) {
        case NORTH:
          jarr[j*jnrow+i] += 10;
          break;
        case EAST:
          jarr[j*jnrow+i] += 3;
          break;
        case WEST:
          jarr[j*jnrow+i] += 6;
          break;
        }
        i--;
        break;
      case SOUTH:
        switch (last) {
        case SOUTH:
          jarr[j*jnrow+i] += 10;
          break;
        case EAST:
          jarr[j*jnrow+i] += 9;
          break;
        case WEST:
          jarr[j*jnrow+i] += 12;
          break;
        }
        i++;
        break;
      case EAST:
        switch (last) {
        case NORTH:
          jarr[j*jnrow+i] += 12;
          break;
        case SOUTH:
          jarr[j*jnrow+i] += 6;
          break;
        case EAST:
          jarr[j*jnrow+i] += 5;
          break;
        }
        j++;
        break;
      case WEST:
        switch (last) {
        case NORTH:
          jarr[j*jnrow+i] += 9;
          break;
        case SOUTH:
          jarr[j*jnrow+i] += 3;
          break;
        case WEST:
          jarr[j*jnrow+i] += 5;
          break;
        }
        j--;
        break;
      }
      last = dir;
      //mexPrintf("Looping:  (%d,%d): %d.\n",i,j,jarr[j*jnrow+i]);
    }
  }

  // output jarr
  if (nlhs > 1) {
    plhs[1] = mxCreateDoubleMatrix(jnrow,jncol, mxREAL);
    arr = mxGetPr(plhs[1]);
    for (j = 0; j <jncol; j++) {
      for (i = 0; i < jnrow; i++) {
        arr[j*jnrow+i] = jarr[j*jnrow+i]&15;
      }
    }
  }

  // prepare output
  plhs[0] = mxCreateDoubleMatrix(nrow,ncol, mxREAL);
  arr = mxGetPr(plhs[0]);
  for (i = 0; i < nrow*ncol; i++) {
    arr[i] = 0;
  }

  // convert joints to seg map
  for (j = 0; j < jncol; j++) {
    inside = false;
    arr[j*nrow] = 0;
    for (i = 0; i < jnrow; ) {
      if (jarr[j*jnrow+i]&1) {
        inside = !inside;
      }
      i++;
      arr[i+j*nrow] = (inside ? 1:0);
    }
    mxAssert(!inside,"Odd number of crossings.");
  }
  for (i = 0; i < nrow; i++) {
    arr[i+jncol*nrow] = 0;
  }

  // free stuff
  mxFree(jarr);
}
// end of main()

//*****************************************************************************
